\ Open Memory Test Language parser
\
\ Copyright 2006 Dennis Ruffer (druffer@worldnet.att.net)
\
\     This program is free software; you can redistribute it and/or modify
\     it under the terms of the GNU General Public License as published by
\     the Free Software Foundation; either version 2 of the License, or
\     (at your option) any later version.
\
\     This program is distributed in the hope that it will be useful,
\     but WITHOUT ANY WARRANTY; without even the implied warranty of
\     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
\     GNU General Public License for more details.
\
\     You should have received a copy of the GNU General Public License
\     along with this program; if not, write to the Free Software
\     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
\
\ Concepts translated from the a lex & yacc front-end to:
\
\ Ad J. van de Goor, Aad Offerman, and Ivo Schanstra:
\ Towards a Uniform Notation for Memory Tests
\ presented at the European Design & Test Conference and Exhibition 96
\
\ In here you can find a predecessor to the language above that is much more
\ limited (more like an extension to the March test notation) and thus
\ much more cleaner.
\
\ Memory Test Verification and Generation
\  Copyright 1995, A. Offerman (offerman@einstein.et.tudelft.nl),
\                  H.I. Schanstra (ivo@duteca.et.tudelft.nl),
\                  Section Computer Architecture & Digital Systems,
\                  Department of Electrical Engeneering,
\                  Delft University of Technology, The Netherlands
\  All rights reserved
\
\ http://www.offerman.net/memory/mtl.tgz

require ../ffl/stc.fs
require ../Vectors.fth
require ../Structs.fth
require Forced_state.fth
require New_state.fth
require Effect_mask_list.fth
require Port.fth
require Memory_type.fth
require Chip_operation.fth

variable Memory_chips   0 Memory_chips !
variable Memory_chip#   0 Memory_chip# !

: mtli_memory_chip ( -- a )   Memory_chips Memory_chip# @ [v] ;

struct: Memory_chip
                    char_*: chips_name
    Forced_state_mask_list: chips_forced_state_mask_list
       New_state_mask_list: chips_new_state_mask_list
          Effect_mask_list: chips_effect_mask_list
               Port_vector: chips_port_vector
                       int: chips_memory_type
                       int: chips_number_of_ports
                       int: chips_total_cells
                       int: chips_height
                       int: chips_width				\ width defined in cells, not in words
                       int: chips_word_width
                       int: chips_width_div_b		\ width_div_b defines width in words, not in cells
\                Real_list: chips_real_state
                State_mask: chips_temp_state
                State_mask: chips_threshold_state
                      cell: chips_address
;struct

: chip_rows ( -- a' )   mtli_memory_chip chips_height ;
: chip_columns ( -- a' )   mtli_memory_chip chips_width_div_b ;
: base_address ( -- a' )   mtli_memory_chip chips_address ;

: chip.reset ( -- ) ;

create MEMORY_CHIP_NAME ," Chip# "

: new_Memory_chip ( address height width type -- )
    Memory_chips @ if
        Memory_chips dup vsize 0 vinsert
    else
        Memory_chip 1 vnew Memory_chips !
    then
    Memory_chips vlast Memory_chip# !
    MEMORY_CHIP_NAME mtli_memory_chip dup >r chips_name !
    ( type ) dup r@ chips_memory_type !  Memory_type# !
    ( width ) r@ chips_width !
    ( height ) r@ chips_height !
    ( address ) r@ chips_address !
    current_memory_type memory_port_type_vector vsize r@ chips_number_of_ports !
    current_memory_type memory_word_width @ r@ chips_word_width !
    r@ chips_width @ r@ chips_word_width @ / r@ chips_width_div_b !
    r@ chips_width @ r@ chips_height @ * r> chips_total_cells !
;

: named_Memory_chip ( address height width type _name -- )
    new_Memory_chip
    >in @  Memory_chip# @ constant
    >in !  bl word count  here mtli_memory_chip chips_name ! $c,
;

create labeled_indices   here 26    cells dup allot -1 fill
create loop_indices      here 26 2* cells dup allot -1 fill

: offset_rows ( index chip_operation -- ) \ changes offset to address
	dup >r  row_index @ ?dup if
		1- cells labeled_indices + @
		2* cells loop_indices + @ dup 0< if
			drop  else  nip  then
	then  r@ row_address @ +
	begin  dup 0< while
		chip_rows @ +
	repeat  chip_rows @ mod
	r> row_address !
;

: offset_columns ( index chip_operation -- ) \ changes offset to address
	dup >r  column_index @ ?dup if
		1- cells labeled_indices + @
		2* 1+ cells loop_indices + @ dup 0< if
			drop  else  nip  then
	then  r@ column_address @ +
	begin  dup 0< while
		chip_columns @ +
	repeat  chip_columns @ mod
	r> column_address !
;

create tmp_chip_operation   Chip_operation allot

defer do_dimensions

: nest_into_dimensions ( chip_operation -- )
	dup row_address @  swap column_address @   1 chip_process# +!
	chip_processing column_exclusion !
	chip_processing row_exclusion !
	do_dimensions  -1 chip_process# +!
;

: data_address ( chip_operation -- data len address len )
	dup >r chip_data vstring
	r@ row_address @ chip_columns @ *
	r@ column_address @ +
	base_address @ +
	r> chip_data vsize
;

: do_chip_operation ( chip_operation -- )
\	dup .chip_operation
	dup chip_core_operation c@ case
		 nest_operation of  nest_into_dimensions  endof
		 read_operation of  data_address compare abort" Failure!"  endof
		write_operation of  data_address rot drop cmove  endof
		dup of  drop  endof
	endcase
;

: row_column_cell ( index2 index1 -- flag )
	2dup loop_indices chip_process# @ 2* cells + 2!
	2dup chip_processing row_exclusion @ dup 0< if
		2drop false  else  =  then
	swap chip_processing column_exclusion @ dup 0< if
		2drop false  else  =  then  and if
		2drop 0 exit ( continue )
	then
	chip_processing chip_operation_list
	dup vsize 0 ?do  dup i [v] tmp_chip_operation Chip_operation move
		rot dup tmp_chip_operation offset_columns
		rot dup tmp_chip_operation offset_rows
		rot tmp_chip_operation do_chip_operation
	loop  drop
	2drop 0
;

: swapped_row_column_cell ( index2 index1 -- flag )   swap row_column_cell ;

: row_cell ( index -- flag )
	dup loop_indices chip_process# @ 2* cells + !
	chip_processing row_exclusion @
	dup 0< if  drop false  else  over =  then  if
		drop 0 exit ( continue )
	then
	chip_processing chip_operation_list
	dup vsize 0 ?do  dup i [v] tmp_chip_operation Chip_operation move
		over tmp_chip_operation offset_rows
		chip_processing column_exclusion @ dup 0< if  drop
		else  tmp_chip_operation offset_columns
		then  tmp_chip_operation do_chip_operation
	loop  drop
	drop 0
;

: column_cell ( index -- flag )
	dup loop_indices chip_process# @ 2* 1+ cells + !
	chip_processing column_exclusion @
	dup 0< if  drop false  else  over =  then  if
		drop 0 exit ( continue )
	then
	chip_processing chip_operation_list
	dup vsize 0 ?do  dup i [v] tmp_chip_operation Chip_operation move
		over tmp_chip_operation offset_columns
		chip_processing column_exclusion @ dup 0< if  drop
		else  tmp_chip_operation offset_rows
		then  tmp_chip_operation do_chip_operation
	loop  drop
	drop 0
;

2variable outer_dimensions
variable outer_increment
variable do_1_cell

: do_1_dimension ( -- )
	outer_dimensions 2@ do
		i do_1_cell @ execute if
			unloop exit
		then  outer_increment @
	+loop
;

2variable inner_dimensions
variable inner_increment
variable do_2_cell

: do_2_dimensions ( -- )
	outer_dimensions 2@ do
		inner_dimensions 2@ do
			j i do_2_cell @ execute if
				unloop unloop exit
			then  inner_increment @
		+loop  outer_increment @
	+loop
;

: (do_dimensions) ( -- )
	chip_processing time_operation @ ?dup if
		0 invert 1 rshift invert xor		\ clear msb
\		cr ." time_operation=" dup .
		1000 min ms
	else
		outer_dimensions 2@ 2>r
		outer_increment @ >r
		do_1_cell @ >r
		chip_processing process_index @ ?dup if
			chip_process# @ labeled_indices rot 1- cells + !
		then
		chip_processing swapped_dimensions c@ if
			chip_rows @ 0
			chip_processing outer_backwards c@ if
				swap chip_processing row_increment @ -
			then
			chip_processing row_increment @
			['] row_cell
		else
			chip_columns @ 0
			chip_processing outer_backwards c@ if
				swap chip_processing column_increment @ -
			then
			chip_processing column_increment @
			['] column_cell
		then  is do_1_cell
		chip_processing outer_backwards c@ if
			negate
		then  outer_increment !
		outer_dimensions 2!
		chip_processing 2_dimensions c@ if
			inner_dimensions 2@ 2>r
			inner_increment @ >r
			do_2_cell @ >r
			chip_processing swapped_dimensions c@ if
				chip_columns @ 0
				chip_processing inner_backwards c@ if
					swap chip_processing column_increment @ -
				then
				chip_processing column_increment @
				['] swapped_row_column_cell
			else
				chip_rows @ 0
				chip_processing inner_backwards c@ if
					swap chip_processing row_increment @ -
				then
				chip_processing row_increment @
				['] row_column_cell
			then  is do_2_cell
			chip_processing inner_backwards c@ if
				negate
			then  inner_increment !
			inner_dimensions 2!
			do_2_dimensions
			r> do_2_cell !
			r> inner_increment !
			2r> inner_dimensions 2!
		else
			do_1_dimension
		then
		r> do_1_cell !
		r> outer_increment !
		2r> outer_dimensions 2!
	then
;
' (do_dimensions) is do_dimensions

: traverse_chip ( -- flag )
	labeled_indices 26    cells -1 fill
	loop_indices    26 2* cells -1 fill
	chip_processes @ dup >r if
		0 chip_process# !
		do_dimensions
	then  r>
;
