\ Open Memory Test Language parser
\
\ Copyright 2006 Dennis Ruffer (druffer@worldnet.att.net)
\
\     This program is free software; you can redistribute it and/or modify
\     it under the terms of the GNU General Public License as published by
\     the Free Software Foundation; either version 2 of the License, or
\     (at your option) any later version.
\
\     This program is distributed in the hope that it will be useful,
\     but WITHOUT ANY WARRANTY; without even the implied warranty of
\     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
\     GNU General Public License for more details.
\
\     You should have received a copy of the GNU General Public License
\     along with this program; if not, write to the Free Software
\     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
\
\ Concepts translated from the a lex & yacc front-end to:
\
\ Ad J. van de Goor, Aad Offerman, and Ivo Schanstra:
\ Towards a Uniform Notation for Memory Tests
\ presented at the European Design & Test Conference and Exhibition 96
\
\ In here you can find a predecessor to the language above that is much more
\ limited (more like an extension to the March test notation) and thus
\ much more cleaner.
\
\ Memory Test Verification and Generation
\  Copyright 1995, A. Offerman (offerman@einstein.et.tudelft.nl),
\                  H.I. Schanstra (ivo@duteca.et.tudelft.nl),
\                  Section Computer Architecture & Digital Systems,
\                  Department of Electrical Engeneering,
\                  Delft University of Technology, The Netherlands
\  All rights reserved
\
\ http://www.offerman.net/memory/mtl.tgz

\ Finally, tests are an important object in the program.

require ../ffl/stc.fs
require ../Vectors.fth
require ../Structs.fth
require Memory_type.fth
require Port.fth
require Fault.fth
require Test_set.fth
require Memory_chip.fth

struct: Test
              char_*: test_name
              char_*: a_test_algorithm
                 int: horizontal_margin
                 int: vertical_margin
                 int: word_width
    Port_type_vector: test_port_type_vector
;struct

: algorithm ( -- a )   current_test a_test_algorithm @ ;

: new_Test ( _name -- a )
    Test_set @ if
        Test_set dup vsize 0 vinsert
    else
        Test 1 vnew Test_set !
    then
    >in @  Test_set vlast dup Test_set# ! constant
    >in !  bl word count  here current_test dup >r test_name ! $c,
    here dup 0 , r> a_test_algorithm !
;

defer verify_chip ( chip -- complete )

variable number_of_places
variable number_to_be_placed
variable number_of_port_mappings
variable max_list_size
variable port_mapping
variable port_mapping_list
variable fault_number_of_ports

: place_port ( port -- )    \ Recursive routine in test.cc
    dup number_to_be_placed @ = if
        port_mapping @ port_mapping_list number_of_port_mappings @ [v] !
        1 cells number_to_be_placed @ vnew port_mapping !
        1 number_of_port_mappings +!
    else
        number_of_places @ 0 ?do
            true ( port free ) over
            ( port ) 0 ?do
                port_mapping i [v] @ j = if
                    drop false
                then
            loop
            ( free ) if
                current_memory_type memory_port_type_vector 0 [v] c@
                current_fault_set fault_port_vector 0 [v] c@
                over dont_care_port =
                over dont_care_port = or 
                -rot = or if
                    port_mapping over [v] i swap !
                    ( port ) dup 1+ recurse \ cr ." I need to recurse place_port here" cr
                    port_mapping over [v] number_of_places @ swap !
                then
            then
        loop
    then  drop
;

variable complete
variable number_mirrored
variable number_of_nbh
variable nbh_counter
variable nbh_counter'
variable nbh_position
variable next_neighborhood
variable next_height
variable next_width
variable hor_low
variable hor_high
variable ver_low
variable ver_high
variable next_nbh
variable chk_height
variable chk_width
variable chk_x
variable chk_y
variable x
variable y
variable eheu
variable x_hi
variable x_lo
variable y_hi
variable y_lo
variable nbh_x
variable nbh_y
variable nbh_width
variable nbh_height
variable number_placed
variable total_width
variable total_height
variable total_cells
variable old_total_cells
variable mca_mapping
variable cell_counter
variable global_x
variable global_y
variable chip
variable num_of_ports
variable mapping
variable nbh_offset
variable temp_port_vector
variable port_counter
variable temp_entry
variable fault_entry
variable old_height
variable row_counter
variable old_width
variable column_counter

: place_neighborhood ( -- )                                                     \ Backtracking procedure positioning neighborhoods
    complete @ if
        number_of_nbh @ ( next_nbh ) 0 ?do
            i next_nbh !                                                        \ We need this in inner loops
            nbh_position i [v] placed @ 0= if                            \ next_neighborhood not yet placed
                number_of_nbh @ ( adjacent_nbh ) -1 ?do
                    i ( adjacent_nbh ) -1 = dup 0= if
                        nbh_position i [v] placed @ or
                    then if                                                     \ adjacent_neighborhood placed
                        current_fault_set fault_neighborhood_list next_nbh @ [v] next_neighborhood !
                        next_neighborhood @ height @ next_height !
                        next_neighborhood @ width @ next_width !
                        i ( adjacent_nbh ) -1 = if                     \ no adjacent nbh. define approximate area for next_nbh
                            next_width @ negate hor_low !
                            current_test word_width @ 1- next_width @ - negate hor_high !
                            0 ver_low !
                            0 ver_high !
                        else                                                    \ define approximate area given a certain adjacent_nbh for next_nbh
                            current_fault_set fault_neighborhood_list i [v] ( adjacent_neighborhood )
                            dup height @ ( adjacent_height ) >r
                            width @ ( adjacent_width ) >r
                            nbh_position i [v] x_relative @
                            current_test word_width @
                            current_test horizontal_margin @ 1+ * 1- -
                            next_width @ - hor_low !
                            nbh_position i [v] x_relative @
                            current_test word_width @
                            current_test horizontal_margin @ 1+ * 1- +
                            r> ( adjacent_width ) + hor_high !
                            nbh_position i [v] y_relative @
                            current_test vertical_margin @ -
                            next_height @ - ver_low !
                            nbh_position i [v] y_relative @
                            current_test vertical_margin @ +
                            r> ( adjacent_height ) + ver_high !
                        then
                        next_width @ 0= if          \ this is a row neighborhood
                            0 hor_low !
                            0 hor_high !
                        then
                        next_height @ 0= if         \ this is a column neighborhood
                            0 ver_low !
                            0 ver_high !
                        then
                        hor_high @ abs 1+ hor_low @ abs ?do \ the abs were added to prevent loops around the number circle
                            i x !
                            ver_high @ abs 1+ ver_low @ abs ?do \ for every (x,y) check constraints and overlap. If OK, then place nbh
                                i y !
                                number_of_nbh @ 0 ?do
                                    i next_nbh @ <> dup if
                                        nbh_position i [v] placed @ and
                                    then if                               \ for every placed neighborhood unequal next_nbh
                                        current_fault_set fault_neighborhood_list i [v] ( chk_neighborhood )
                                        dup height @ chk_height !
                                        width @ chk_width !
                                        nbh_position i [v] x_relative @ chk_x !
                                        nbh_position i [v] y_relative @ chk_y !
                                        next_neighborhood @ neighbor_overlap_constraint_list i [v] @
                                        overlap =
                                        x @ chk_x @ next_width @ - <= or
                                        x @ chk_x @ chk_width @ + < and
                                        y @ chk_y @ next_height @ - > and
                                        y @ chk_y @ chk_height @ + < and if     \ (x,y) does not overlap chk_nbh or overlap is allowed
                                            false eheu !                        \ now check horizontal and vertical constraints
                                            next_neighborhood @ horizontal_position_constraint_list i [v] c@ ( next_hor_constraint )
                                            dup low = if  x @ chk_x @ >= eheu !  else
                                            dup equal = if  x @ next_width @ + chk_x @ <=
                                                               x @ chk_x @ chk_width @ + >=
                                                               or eheu !  else
                                            dup high = if  x @ next_width @ + chk_x @ chk_width @ +
                                                              <= eheu !  else
                                            dup low_equal = if  x @ chk_x @ chk_width @ +
                                                                   >= eheu !  else
                                            dup low_high = if  x @ chk_x @ >=
                                                                  x @ next_width @ +
                                                                  chk_x @ chk_width @ <=
                                                                  and eheu !  else
                                            dup equal_high = if  x @ next_width @ + chk_x @ <= eheu !  else
                                            dup no_constraint = if  else
                                            dup not_low = if  x @ chk_x @ < eheu !  else
                                            dup not_equal = if  x @ next_width @ + chk_x @ >
                                                                   x @ chk_x @ chk_width @ + < and eheu !  else
                                            dup not_high = if  x @ next_width @ + chk_x @ chk_width + > eheu !  else
                                            dup not_low_equal = if  x @ chk_x @ chk_width @ + < eheu !  else
                                            dup not_low_high = if  x @ chk_x @ <
                                                                      x @ next_width @ + chk_x @ chk_width @ + >
                                                                      or eheu !  else
                                            dup not_equal_high = if  x @ next_width @ chk_x @ > eheu !  else
                                            dup low_not_high = if  x @ chk_x @ >=
                                                                      x @ next_width @ + chk_x @ chk_width @ + >
                                                                      or eheu !  else
                                            dup equal_not_low = if  x @ chk_x @ chk_width @ + >=
                                                                       x @ chk_x @ < or eheu !  else
                                            dup equal_not_high = if  x @ next_width @ + chk_x @ <=
                                                                        x @ next_width @ + chk_x @ chk_width @ + >
                                                                        or eheu !  else
                                            dup high_not_low = if  x @ next_width @ + chk_x @ chk_width @ + <=
                                                                      x @ chk_x < or eheu !
                                            true abort" illegal position constraint in place_nbh hor"
                                            then then then then then then then then then
                                            then then then then then then then then drop
                                            eheu @ 0= if
                                                next_neighborhood @ vertical_position_constraint_list i [v] c@ ( next_ver_constraint )
                                                dup low = if  y @ chk_y @ >= eheu !  else
                                                dup equal = if  y @ next_height @ + chk_y @ <=
                                                                   y @ chk_y @ chk_height @ + >=
                                                                   or eheu !  else
                                                dup high = if  y @ next_height @ + chk_y @ chk_height @ +
                                                                  <= eheu !  else
                                                dup low_equal = if  y @ chk_y @ chk_height @ +
                                                                       >= eheu !  else
                                                dup low_high = if  y @ chk_y @ >=
                                                                      y @ next_height @ +
                                                                      chk_y @ chk_height @ <=
                                                                      and eheu !  else
                                                dup equal_high = if  y @ next_height @ + chk_y @ <= eheu !  else
                                                dup no_constraint = if  else
                                                dup not_low = if  y @ chk_y @ < eheu !  else
                                                dup not_equal = if  y @ next_height @ + chk_y @ >
                                                                       y @ chk_y @ chk_height @ + < and eheu !  else
                                                dup not_high = if  y @ next_height @ + chk_y @ chk_height + > eheu !  else
                                                dup not_low_equal = if  y @ chk_y @ chk_height @ + < eheu !  else
                                                dup not_low_high = if  y @ chk_y @ <
                                                                          y @ next_height @ + chk_y @ chk_height @ + >
                                                                          or eheu !  else
                                                dup not_equal_high = if  y @ next_height @ chk_y @ > eheu !  else
                                                dup low_not_high = if  y @ chk_y @ >=
                                                                          y @ next_height @ + chk_y @ chk_height @ + >
                                                                          or eheu !  else
                                                dup equal_not_low = if  y @ chk_y @ chk_height @ + >=
                                                                           y @ chk_y @ < or eheu !  else
                                                dup equal_not_high = if  y @ next_height @ + chk_y @ <=
                                                                            y @ next_height @ + chk_y @ chk_height @ + >
                                                                            or eheu !  else
                                                dup high_not_low = if  y @ next_height @ + chk_y @ chk_height @ + <=
                                                                          y @ chk_y < or eheu !
                                                true abort" illegal position constraint in place_nbh ver"
                                                then then then then then then then then then
                                                then then then then then then then then drop
                                                eheu @ 0= if                                    \ check word position constraint (for B-bit wide memories)
	x @ ( x_offset )                            \ x_offset now gives x of nbh's leftmost column
                                                    dup word_width @ / dup 0 < if  negate  then
                                                    2 + word_width @ * +                        \ x_offset now is positive
                                                    word_width @ mod >r                         \ x_offset now between 0 and word_width-1
                                                    next_neighborhood @ word_position_constraint @
                                                    next_neighborhood @ shift_mask r> [v] @ zero = and if
                                                        true eheu !                             \ check identity mask for all ports / decoders and sensebank
                                                    then
                                                    eheu @ 0= if                                \ eheu = TRUE if position is rejected
                                                        \ this (x,y) is a fine place for this neighborhood. It can now be added.
                                                        x @ nbh_position next_nbh @ [v] x_relative !
                                                        y @ nbh_position next_nbh @ [v] y_relative !
                                                        true nbh_position next_nbh @ [v] placed !
                                                        1 number_placed +!
                                                        number_placed @ number_of_nbh @ < if
                                                            recurse \ cr ." I need to recurse place_neighborhood here" cr
                                                        else
                                                            word_width @ negate x_hi !
                                                            word_width @ x_lo !
                                                            0 y_hi !
                                                            0 y_lo !
                                                            number_of_nbh 0 ?do
                                                                nbh_position i [v] x_relative @ nbh_x !
                                                                current_fault_set fault_neighborhood_list i [v] width @ nbh_width !
                                                                nbh_x @ x_lo @ < if
                                                                    nbh_x @ x_lo !
                                                                then
                                                                nbh_x @ nbh_width @ + x_hi @ > if
                                                                    nbh_x @ nbh_width @ + x_hi !
                                                                then
                                                                nbh_position i [v] y @ nbh_y !
                                                                current_fault_set fault_neighborhood_list i [v] height @ nbh_height !
                                                                nbh_y @ y_lo @ < if
                                                                    nbh_y @ y_lo !
                                                                then
                                                                nbh_y @ nbh_height @ + y_hi @ > if
                                                                    nbh_y @ nbh_height @ + y_hi !
                                                                then
                                                            loop
                                                            \ expand entire width to word width with correct boundaries
                                                            -1 x_hi +!
                                                            x_lo @ dup abs = if
                                                                x_lo dup @ dup word_width @ mod - swap !
                                                            else
                                                                x_lo @ negate word_width @ mod if
                                                                    word_width @ x_lo @ negate - word_width @ mod x_lo !
                                                                then
                                                            then
                                                            x_hi @ dup abs = if
                                                                x_hi @ 1+ word_width @ mod if
                                                                    word_width @ x_hi @ 1+ - word_width @ mod x_hi !
                                                                then
                                                            else
                                                                x_hi dup @ dup 1- negate word_width @ mod + swap !
                                                            then
                                                            1 x_hi +!
                                                            x_hi @ x_lo @ - dup total_width !
                                                            y_hi @ y_lo @ - dup total_height !
                                                            * total_cells !
                                                            \ heigth and width determined. build memory_chip.
                                                            \ First calculate mapping.
                                                            current_fault_set fault_cell_size @ old_total_cells !
                                                            1 cells old_total_cells @ vnew mca_mapping !
                                                            0 cell_counter !
                                                            old_total_cells @ 0 ?do
                                                                i nbh_counter' !
                                                                current_fault_set fault_neighborhood_list i [v] width @ nbh_width !
                                                                current_fault_set fault_neighborhood_list i [v] height @ nbh_height !
                                                                nbh_width @ 0 ?do
                                                                    nbh_height @ 0 ?do
                                                                        nbh_position nbh_counter' @ [v] x @ j + x_lo @ - global_x !
                                                                        nbh_position nbh_counter' @ [v] y @ i + y_lo @ - global_y !
                                                                        global_x @ total_width @ global_y @ * + mca_mapping cell_counter @ [v] !
                                                                        1 cell_counter +!
                                                                    loop
                                                                loop
                                                            loop
                                                            \ MCA mapping calculated.
	\ Now build MCA part of memory_chip & expand all masks.
                                                            0 total_height @ total_width @ Memory_type# @ new_Memory_chip
                                                            total_cells @ mca_mapping forced_state_mask_list.expand
                                                                mtli_memory_chip chips_forced_state_mask_list !
                                                            total_cells @ mca_mapping new_state_mask_list.expand
                                                                mtli_memory_chip chips_new_state_mask_list !
                                                            total_cells @ mca_mapping effect_mask_list.expand
                                                                mtli_memory_chip chips_effect_mask_list !
	\ build port_vector part of memory_chip
                                                            current_fault_set fault_port_vector vsize num_of_ports !
                                                            old_total_cells @ nbh_offset !
                                                            1 cells 0 vnew temp_port_vector !
                                                            num_of_ports @ ( port_counter ) 0 do
                                                                i port_counter !
                                                                Port_struct 1 vnew temp_entry !
                                                                current_fault_set fault_port_vector port_counter @ [v] fault_entry !
                                                                fault_entry @ ports_type @ temp_entry @ ports_type !
                                                                fault_entry @ stuck_open_mask total_cells @ mapping state_mask.expand temp_entry @ stuck_open_mask !

                                                                \ build mapping for row decoder and expand row decoder
                                                                fault_entry @ row_decoder decoder_connectivity_matrix vsize old_height !
                                                                1 cells old_height @ vnew mapping !
                                                                nbh_counter @ nbh_offset +!
                                                                0 row_counter !
                                                                fault_entry @ row_decoder decoder_identity
                                                                a_neighborhood_list vsize 0 do
                                                                    fault_entry @ row_decoder decoder_identity
                                                                    a_neighborhood_list i [v]
                                                                    height @ 0 do
                                                                        nbh_position nbh_offset @ nbh_counter @ + [v] y_relative @
                                                                        nbh_y @ + y_lo @ - mapping row_counter @ [v] !
                                                                        1 row_counter +!
                                                                    loop
                                                                loop
                                                                fault_entry @ row_decoder total_cells @ mapping decoder.expand
                                                                temp_entry @ row_decoder !
                                                                \ check identities row decoder
                                                                old_height @ 0 do
                                                                    fault_entry @ row_decoder decoder_identity
                                                                    a_identity_mask i [v] @ -1 <> if
                                                                        mca_mapping i [v] @ total_width @ /
                                                                        mapping j [v] @ <> if
                                                                            true eheu !
                                                                        then
                                                                    then
                                                                loop
                                                                mapping vdelete

                                                                \ build mapping for column decoder and expand column decoder
                                                                fault_entry @ column_decoder decoder_connectivity_matrix vsize old_width !
                                                                1 cells old_width @ vnew mapping !
                                                                nbh_counter @ nbh_offset +!
                                                                0 column_counter !
                                                                fault_entry @ column_decoder decoder_identity
                                                                a_neighborhood_list vsize 0 do
                                                                    fault_entry @ column_decoder decoder_identity
                                                                    a_neighborhood_list i [v]
                                                                    width @ word_width @ / 0 do
                                                                        nbh_position nbh_offset @ nbh_counter @ + [v] x_relative @ word_width @ /
                                                                        nbh_x @ + x_lo @ word_width @ / - mapping column_counter @ [v] !
                                                                        1 column_counter +!
                                                                    loop
                                                                loop
                                                                fault_entry @ column_decoder total_cells @ mapping decoder.expand
                                                                temp_entry @ column_decoder !
                                                                \ check identities column decoder
                                                                old_width @ 0 do
                                                                    fault_entry @ column_decoder decoder_identity
                                                                    a_identity_mask i [v] @ -1 <> if
                                                                        mca_mapping i [v] @ total_width @ mod
                                                                        mapping j [v] @ <> if
                                                                            true eheu !
                                                                        then
                                                                    then
                                                                loop
                                                                mapping vdelete

                                                                \ build mapping for sense bank and expand sense bank
                                                                fault_entry @ ports_sense_bank sense_bank_senseamp_list vsize old_width !
                                                                1 cells old_width @ vnew mapping !
                                                                nbh_counter @ nbh_offset +!
                                                                0 column_counter !
                                                                fault_entry @ ports_sense_bank sense_bank_identity
                                                                a_neighborhood_list vsize 0 do
                                                                    fault_entry @ ports_sense_bank sense_bank_identity
                                                                    a_neighborhood_list i [v]
                                                                    width @ 0 do
                                                                        nbh_position nbh_offset @ nbh_counter @ + [v] x_relative @
                                                                        nbh_x @ + x_lo @ - mapping column_counter @ [v] !
                                                                        1 column_counter +!
                                                                    loop
                                                                loop
                                                                fault_entry @ ports_sense_bank total_cells @ mapping sense_bank.expand
                                                                temp_entry @ ports_sense_bank !
                                                                \ check identities sense bank
                                                                old_width @ 0 do
                                                                    fault_entry @ ports_sense_bank sense_bank_identity
                                                                    a_identity_mask i [v] @ -1 <> if
                                                                        mca_mapping i [v] @ total_width @ mod
                                                                        mapping j [v] @ <> if
                                                                            true eheu !
                                                                        then
                                                                    then
                                                                loop
                                                                mapping vdelete

                                                                \ insert this port in temp_port_vector
                                                                temp_port_vector port_counter @ temp_entry vinsert
                                                            loop
                                                            mca_mapping vdelete
                                                            eheu @ 0= if
                                                                \ no identity violation. continue with assigning ports expansion.
                                                                
                                                                \ use port_mapping_vector and number_of_port_mappings
                                                                \ first build all ports as good ports
                                                                \ then overwrite some ports with faulty ports
                                                                
                                                                \ port expansion
                                                                Memory_chip# @ verify_chip complete !
                                                            then
                                                        then
                                                        \ remove neighborhood for backtracking
                                                        false nbh_position next_nbh @ [v] placed !
                                                        -1 number_placed +!
                                                    then
                                                then
                                            then
                                        then
                                    then
                                loop
                            loop
                        loop
                    then
                loop
            then
        loop
    then
;

defer mirror_nbh ( -- )   \ Recursive procedure mirroring neighborhoods

: mirror_nbh_diagonal ( -- )
	2 ( nbh_dmc ) 0 do
		i ( nbh_dmc ) 1 = if
			current_fault_set fault_port_vector nbh_counter @ [v]
			neighbor_mirror_constraint diagonal c@ not_allowed = if
				leave
			else
				nbh_counter @ d_mirror_nbh
			then
		then
		1 nbh_counter +!
		nbh_counter @ number_of_nbh @ < if
			mirror_nbh \ recurse through deferred word
		else
			Relative_position number_of_nbh @ vnew nbh_position !
			number_of_nbh @ 0 ?do
				0 nbh_position i [v] x_relative !
				0 nbh_position i [v] y_relative !
				false nbh_position i [v] placed !
			loop
			0 number_placed !
			place_neighborhood
			nbh_position vdelete
		then
		-1 nbh_counter +!
		i ( nbh_dmc ) 1 = if
			nbh_counter @ d_mirror_nbh
		then
	loop
;

: mirror_nbh_vertical ( -- )
	2 ( nbh_vmc ) 0 do
		i ( nbh_vmc ) 1 = if
			current_fault_set fault_port_vector nbh_counter @ [v]
			neighbor_mirror_constraint vertical c@ not_allowed = if
				leave
			else
				nbh_counter @ v_mirror_nbh
			then
		then
		mirror_nbh_diagonal
		i ( nbh_vmc ) 1 = if
			nbh_counter @ v_mirror_nbh
		then
	loop
;

: mirror_nbh_horizontal ( -- )
	2 ( nbh_hmc ) 0 do
		i ( nbh_hmc ) 1 = if
			current_fault_set fault_port_vector nbh_counter @ [v]
			neighbor_mirror_constraint horizontal c@ not_allowed = if
				leave
			else
				nbh_counter @ h_mirror_nbh
			then
		then
		mirror_nbh_vertical
		i ( nbh_hmc ) 1 = if
			nbh_counter @ h_mirror_nbh
		then
	loop
;

: (mirror_nbh) ( -- )   \ Recursive procedure mirroring neighborhoods
    complete @ if
        mirror_nbh_horizontal
    then
;
' (mirror_nbh) is mirror_nbh

defer mirror_port ( -- )   \ Recursive procedure mirroring ports

: expand_diagonal_mirroring ( -- )   \ expand diagonal MCA mirroring
	2 ( array_dmc ) 0 do
		i ( array_dmc ) 1 = if
			current_fault_set fault_array_mirror_constraint
			diagonal c@ not_allowed = if
				leave
			else
				d_mirror_mca
			then
		then
		i ( array_dmc ) 1 = if
			d_mirror_mca
		then
	loop
;

: expand_vertical_mirroring ( -- )
	2 ( array_vmc ) 0 do
		i ( array_vmc ) 1 = if
			current_fault_set fault_array_mirror_constraint
			vertical c@ not_allowed = if
				leave
			else
				current_fault_set fault_neighborhood_list v_complement
			then
		then
		expand_diagonal_mirroring
		\ call procedure expand diagonal nbh mirroring
		current_fault_set fault_neighborhood_list vsize number_of_nbh !
		0 nbh_counter !
		mirror_nbh
		i ( array_vmc ) 1 = if
			current_fault_set fault_neighborhood_list v_complement
		then
	loop
;

: expand_horizontal_mirroring ( -- )   \ expand MCA mirroring through constraints 
	\ array_hmc = array_horizontal_mirror_counter
	2 ( array_hmc ) 0 do
		i ( array_hmc ) 1 = if
			current_fault_set fault_array_mirror_constraint
			horizontal c@ not_allowed = if
				leave
			else
				current_fault_set fault_neighborhood_list h_complement
			then
		then
		expand_vertical_mirroring
		i ( array_hmc ) 1 = if
			current_fault_set fault_neighborhood_list h_complement
		then
	loop
;

: mirror_sense_banks ( -- )
	2 ( senseb_mc ) 0 do
		i ( senseb_mc ) 1 = if
	        current_fault_set fault_port_vector number_mirrored @ [v]
			ports_sense_bank sense_bank_identity
			a_mirror_constraint c@ not_allowed = if
				leave
			else
		        current_fault_set fault_port_vector number_mirrored @ [v]
				dup ports_sense_bank sense_bank_identity
				a_neighborhood_list h_complement
			then
		then
        current_fault_set fault_port_vector number_mirrored @ [v] c@    \ ????
		current_fault_set fault_port_vector number_mirrored @ [v] c!    \ ????
		1 number_mirrored +!
		number_mirrored @ fault_number_of_ports @ < if
			mirror_port \ recurse through deferred word
		else
			\ all ports mirrored
			expand_horizontal_mirroring
		then
		-1 number_mirrored +!
		i ( senseb_mc ) 1 = if
	        current_fault_set fault_port_vector number_mirrored @ [v]
			ports_sense_bank sense_bank_identity
			a_neighborhood_list h_complement
		then
	loop
;

: mirror_columns ( -- )
	2 ( coldec_mc ) 0 do
		i ( coldec_mc ) 1 = if
	        current_fault_set fault_port_vector number_mirrored @ [v]
			column_decoder decoder_identity
			a_mirror_constraint c@ not_allowed = if
				leave
			else
		        current_fault_set fault_port_vector number_mirrored @ [v]
				column_decoder decoder_identity
				a_neighborhood_list h_complement
			then
		then
		mirror_sense_banks
		i ( coldec_mc ) 1 = if
	        current_fault_set fault_port_vector number_mirrored @ [v]
			column_decoder decoder_identity
			a_neighborhood_list h_complement
		then
	loop
;

: mirror_rows ( -- )
	2 ( rowdec_mc ) 0 do
		i ( rowdec_mc ) 1 = if
	        current_fault_set fault_port_vector number_mirrored @ [v]
			row_decoder decoder_identity
			a_mirror_constraint c@ not_allowed = if
				leave
			else
		        current_fault_set fault_port_vector number_mirrored @ [v]
				row_decoder decoder_identity
				a_neighborhood_list v_complement
			then
		then
		mirror_columns
		i ( rowdec_mc ) 1 = if
            current_fault_set fault_port_vector number_mirrored @ [v]
			row_decoder decoder_identity
			a_neighborhood_list v_complement
		then
	loop
;

: (mirror_port) ( -- )   \ Recursive procedure mirroring ports
    complete @ if
        mirror_rows
    then
;
' (mirror_port) is mirror_port

: verify_completeness ( -- complete )           \ Verify test for completeness with fault
    Fault_set# @ >r
    true complete !
    Failt_set.size 0 do
        i Fault_set# !                          \ build port_mapping
        current_memory_type memory_port_type_vector vsize number_of_places !
        current_fault_set fault_port_vector vsize dup number_to_be_placed !
        1 cells swap vnew port_mapping !
        1 max_list_size !
        number_to_be_placed @ 0 ?do
            max_list_size dup @ number_of_places @ i - * swap !
        loop
max_list_size dup @ 4 * swap !					\ Why 4 times as many ???
        1 cells max_list_size @ vnew port_mapping_list !
        max_list_size @ 0 ?do
            0 port_mapping_list i [v] !
        loop
        0 number_of_port_mappings !
        number_to_be_placed @ 0 > if
            0 place_port
            number_of_port_mappings @ 0= if
                cr ." Didn't place any ports"
                false complete !
                leave
            then
        then                                    \ put port nbh's in same list as mca nbh's
        current_fault_set fault_port_vector vsize
        dup fault_number_of_ports ! 0 do
            i 0 row_decoder         decoder_identity append_fault_neighbor
            i 0 column_decoder      decoder_identity append_fault_neighbor
            i 0 ports_sense_bank sense_bank_identity append_fault_neighbor
        loop
    loop
                                                \ expand non_nbh mirroring through constraints 
                                                \ start with expanding port_vector
    complete @ if
        0 number_mirrored !
        mirror_port
    then
    complete @
    r> Fault_set# !
;
