\ Open Memory Test Language vectors
\
\ Copyright 2006 Dennis Ruffer (druffer@worldnet.att.net)
\
\     This program is free software; you can redistribute it and/or modify
\     it under the terms of the GNU General Public License as published by
\     the Free Software Foundation; either version 2 of the License, or
\     (at your option) any later version.
\
\     This program is distributed in the hope that it will be useful,
\     but WITHOUT ANY WARRANTY; without even the implied warranty of
\     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
\     GNU General Public License for more details.
\
\     You should have received a copy of the GNU General Public License
\     along with this program; if not, write to the Free Software
\     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
\
\ Concepts referenced in the lex & yacc front-end to:
\
\ Ad J. van de Goor, Aad Offerman, and Ivo Schanstra:
\ Towards a Uniform Notation for Memory Tests
\ presented at the European Design & Test Conference and Exhibition 96
\
\ There are no source definitions to these words, only references.
\
\ Memory Test Verification and Generation
\  Copyright 1995, A. Offerman (offerman@einstein.et.tudelft.nl),
\                  H.I. Schanstra (ivo@duteca.et.tudelft.nl),
\                  Section Computer Architecture & Digital Systems,
\                  Department of Electrical Engeneering,
\                  Delft University of Technology, The Netherlands
\  All rights reserved
\
\ http://www.offerman.net/memory/mtl.tgz

warnings @ warnings off

: append ( from len to -- )   2dup >r >r  count + swap move  r> r@ c@ + r> c! ;
: place ( from len to -- )   0 over c! swap 255 min swap append ;

: $c, ( str len -- )   here  over 1+ allot  place ;

: @+ ( addr -- str len )   dup cell+ swap @ ;

: +$, ( str len dest -- )
    dup @+ + here <> abort" can't concatenate strings"
    over allot  2dup >r >r  @+ + swap move  r> r@ @ + r> !
;

: ,"      ( _"..." -- )   [char] " word count      $c, ;
: +" ( addr _"..." -- )   [char] " word count rot +$, ; 

warnings !

: vdim ( size lim -- size lim dim )   2dup * 2 cells + ;

: vnew ( size n -- a )
    vdim dup >r allocate throw
    dup r> erase dup >r 2!
    r> 2 cells +
;

: vdelete ( a -- )
    dup @ ?dup if  2 cells - free throw  then
    0 swap !
;

: [v] ( a n -- a' )
    swap @ ?dup 0= abort" Vector does not exist"
    dup 2 cells - 2@ >r
    rot dup r> >=
	over 0 < or abort" [v] index out of bounds"
    * +
;

: vsize ( a -- n )   @ dup if 2 cells - @ then ;
: vlast ( a -- n )   vsize 1- ;

: vinsert ( a n x -- )          \ only adds element to the end
    rot dup >r @                \ n x (a)
    2 cells - dup 2@            \ n x (a)-2cells size lim
    1+ vdim                     \ n x (a)-2cells size lim+1 dim
    -rot 2swap resize throw     \ n x size lim+1 anew
    dup >r 2! r> 2 cells +      \ n x anew+2cells
    r@ ! r> dup vlast [v] c!    \ n x-> seeds new element
    drop                        \ discard the index
;

: vextract ( a n -- )           \ only removes the last element
    drop dup >r @               \ (a)
    2 cells - dup 2@            \ (a)-2cells size lim
    1- 0 max vdim               \ (a)-2cells size lim-1 dim
    -rot 2swap resize throw     \ size lim-1 anew
    dup >r 2! r> 2 cells +      \ anew+2cells
    r> !
;

: atos ( str len -- a )   1 over vnew dup >r swap move r> ;

: atod ( str len -- n )   0 0 2swap >number 2drop ;

: atoi ( str len -- n )   atod drop ;

: vstring ( a -- str len )   dup @ swap vsize ;

: vtoi ( a -- n )   vstring atoi ;

\ ***	The following are used to fetch data that is in known Endian format.  E.g.
\ *** in file system structures or network packets.  These words work on un-aligned
\ *** entities.

: 1c!-le ( x a n -- )   BEGIN ?DUP WHILE
	1- ROT DUP 8 RSHIFT SWAP 2OVER DROP C! ROT 1+ ROT REPEAT 2DROP ;
: 1c!-be ( x a n -- )   BEGIN ?DUP WHILE
	1- ROT DUP 8 RSHIFT SWAP 2OVER +    C! ROT    ROT REPEAT 2DROP ;
: 1c@-le ( a n -- x )   0 SWAP BEGIN ?DUP WHILE
 	1- ROT 2DUP +      C@ >R ROT 8 LSHIFT R> + ROT REPEAT NIP ;
: 1c@-be ( a n -- x )   0 SWAP BEGIN ?DUP WHILE
	1- ROT DUP 1+ SWAP C@ >R ROT 8 LSHIFT R> + ROT REPEAT NIP ;

\ ***	Which order is little-endian multi-cell data in?
\ *** I assume the same order as big-endian for now!
: c!-le ( ... a n -- )   1 CELLS /MOD SWAP >R  ?DUP
	IF  0 DO  SWAP OVER 1 CELLS 1c!-le CELL+  LOOP
	THEN  R> ?DUP IF  1c!-le  ELSE DROP THEN ;
: c!-be ( ... a n -- )   1 CELLS /MOD SWAP >R  ?DUP
	IF  0 DO  SWAP OVER 1 CELLS 1c!-be CELL+  LOOP
	THEN  R> ?DUP IF  1c!-be  ELSE DROP THEN ;
: c@-le ( a n -- ... )
	1 CELLS /MOD ROT OVER CELLS + -ROT SWAP ?DUP
	IF  ROT DUP ROT 1c@-le -ROT SWAP  THEN  ?DUP
	IF  0 DO  1 CELLS - DUP 1 CELLS 1c@-le SWAP
	LOOP  THEN  DROP ;
: c@-be ( a n -- ... )
	1 CELLS /MOD ROT OVER CELLS + -ROT SWAP ?DUP
	IF  ROT DUP ROT 1c@-be -ROT SWAP  THEN  ?DUP
	IF  0 DO  1 CELLS - DUP 1 CELLS 1c@-be SWAP
	LOOP  THEN  DROP ;
