\ MTL Parser based on gforth's Gray version 5
\
\ Copyright 2006 Dennis Ruffer (druffer@worldnet.att.net)
\
\     This program is free software; you can redistribute it and/or modify
\     it under the terms of the GNU General Public License as published by
\     the Free Software Foundation; either version 2 of the License, or
\     (at your option) any later version.
\
\     This program is distributed in the hope that it will be useful,
\     but WITHOUT ANY WARRANTY; without even the implied warranty of
\     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
\     GNU General Public License for more details.
\
\     You should have received a copy of the GNU General Public License
\     along with this program; if not, write to the Free Software
\     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
\

wordlist constant mtla_wordlist
get-current ( * ) mtla_wordlist set-current
get-order mtla_wordlist swap 1+ set-order

\ parsing stacks

40 stack integers
30 stack bits_per_words
30 stack vertical_periods
30 stack horizontal_periods
30 stack port_type_vectors
30 stack numbers_of_ports

: mtla_test_insert ( -- )
	horizontal_periods dup pull mtla_horizontal_period ! clear? 0= abort" horizontal_periods left"
	vertical_periods dup pull mtla_vertical_period ! clear? 0= abort" vertical_periods left"
	bits_per_words dup pull mtla_bits_per_word ! clear? 0= abort" bits_per_words left"
	begin  mtla_port_type_vector vsize while
		mtla_port_type_vector 0 vextract
	repeat
	port_type_vectors dup @
	numbers_of_ports dup pull 0 ?do
		over i [v] c@ case
			      read_port of  mtla_port_type_vector i read_port       vinsert  endof
			     write_port of  mtla_port_type_vector i write_port      vinsert  endof
			read_write_port of  mtla_port_type_vector i read_write_port vinsert  endof
			 dont_care_port of  mtla_port_type_vector i dont_care_port  vinsert  endof
			true abort" MTL internal error: port_type not recognized"
		endcase
	loop  nip
	clear? 0= abort" numbers_of_ports left"
	dup pop clear? 0= abort" port_type_vectors left"
;

: set_bits_per_word ( -- )
	bits_per_words pull dup
	bits_per_words top = if  drop
	else  bits_per_words top 1 = if
			bits_per_words dup pop push
		else
			1 <> abort" MTL attributes error: bits per word conflict"
		then
	then
;

: set_hv_bits_per_word ( -- )
	horizontal_periods dup pull over pull max swap push
	vertical_periods dup pull over pull max swap push
	set_bits_per_word
;

variable tmp1_port_type_vector
variable tmp2_port_type_vector

: set_port_type_vectors ( n -- )
	port_type_vectors pull tmp1_port_type_vector !
	port_type_vectors pull tmp2_port_type_vector !
	1 chars numbers_of_ports top dup >r vnew
	port_type_vectors push
	r> 0 ?do
		tmp1_port_type_vector r@ [v] c@ read_write_port =
		tmp2_port_type_vector r@ [v] c@ read_write_port = or if
			read_write_port port_type_vectors @ r@ [v] c!
		else
			tmp1_port_type_vector r@ [v] c@
			tmp2_port_type_vector r@ [v] c@ = if
				tmp2_port_type_vector r@ [v] c@
				port_type_vectors @ r@ [v] c!
			else
				tmp1_port_type_vector r@ [v] c@ dont_care_port = if
					tmp2_port_type_vector r@ [v] c@
					port_type_vectors @ r@ [v] c!
				else
					tmp2_port_type_vector r@ [v] c@ dont_care_port = if
						tmp1_port_type_vector r@ [v] c@
						port_type_vectors @ r@ [v] c!
					else
						read_write_port port_type_vectors @ r@ [v] c!
					then
				then
			then
		then
	loop
	tmp1_port_type_vector vdelete
	tmp2_port_type_vector vdelete
;

: set_port_type_vector ( -- )
	set_hv_bits_per_word
	numbers_of_ports pull dup
	numbers_of_ports top = if  drop
		set_port_type_vectors
	else
		numbers_of_ports top 0= if
			port_type_vectors pull
			port_type_vectors dup @ vdelete dup pop push
			numbers_of_ports dup pop push
		else
			abort" MTL attributes error: number of ports conflict"
			port_type_vectors dup @ vdelete pop
		then
	then
;

: add_port_type_vector ( -- )
	set_hv_bits_per_word
	port_type_vectors pull tmp1_port_type_vector !
	port_type_vectors pull tmp2_port_type_vector !
	numbers_of_ports pull dup 1+ dup
	1 chars over vnew
	port_type_vectors push
	numbers_of_ports dup pop push
	tmp2_port_type_vector 0 [v] c@
	port_type_vectors @ 0 [v] !
	0 ?do
		tmp1_port_type_vector i [v] c@
		port_type_vectors @ i 1+ [v] c!
	loop
	tmp1_port_type_vector vdelete
	tmp2_port_type_vector vdelete
;

: add_a_port_type_vector ( type bits -- )
	( bits )		bits_per_words push
	1 chars 1 vnew	port_type_vectors push
	1				numbers_of_ports push
	( type )		port_type_vectors @ 0 [v] c!
;

: add_string_port_type_vector ( type -- )
	bits_per_words pull add_a_port_type_vector
;

: add_hv_port_type_vector ( horizontal vertical type bits -- )
	2swap vertical_periods push
	horizontal_periods push
	add_a_port_type_vector
;

: add_hv_string_port_type_vector ( horizontal vertical type -- )
	bits_per_words pull	add_hv_port_type_vector
;

: add_last_port_type_vector ( type -- )
	horizontal_periods pull
	vertical_periods pull
	rot add_hv_string_port_type_vector
;

: push_defaults ( -- )
	1 horizontal_periods push
	1 vertical_periods push
	1 bits_per_words push
	0 port_type_vectors push
	0 numbers_of_ports push
;

: mtla_init ( -- )
	integers clear
	bits_per_words clear
	vertical_periods clear
	horizontal_periods clear
	begin  port_type_vectors clear? 0= while
		port_type_vectors dup @ vdelete pop
	repeat
	numbers_of_ports clear
;

\ non-terminals

nonterminal <sub_march_element_list>

\ parser

(( <unsigned_integer> {{ ident 2@ atoi integers push }} )) <- <integer>
(( <unsigned_integer> {{ ident 2@ bits_per_words push drop }}  )) <- <data>

(( <t> (( <integer> )) ?? )) <- <time_operation>

(( <time_operation> )) <- <global_march_element>

(( <integer> )) <- <row_offset>
(( <integer> )) <- <column_offset>

(( <[> <row_offset> <,> <column_offset> <]> )) <- <location>

(( <r> <location> <data> {{  read_port add_string_port_type_vector }}
|| <w> <location> <data> {{ write_port add_string_port_type_vector }}
)) <- <local_operation>

(( <integer> {{ integers pull vertical_periods push }} )) <- <number_of_rows>
(( <integer> {{ integers pull horizontal_periods push }} )) <- <number_of_columns>

(( <[> <number_of_rows> <,> <number_of_columns> <]> )) <- <tile_size>

(( <local_operation> (( <,> <local_operation>
{{ set_bits_per_word 1 numbers_of_ports dup pop dup pop push
   set_port_type_vectors }}
)) ** )) <- <local_operation_list>

(( <UPDOWN> || <DOWN> || <UP> )) <- <_1_dimensional_addressing_direction>

(( <EASTWEST> || <EAST> || <WEST> )) <- <row_addressing_direction>

(( <NORTHSOUTH> || <SOUTH> || <NORTH> )) <- <column_addressing_direction>

(( <a> || <b> || <c> || <d> || <e> || <f> || <g> || <h> || <i> || <j> || <k> 
|| <l> || <m> || <n> || <o> || <p> || <q> || <r> || <s> || <t> || <u> || <v> 
|| <w> || <x> || <y> || <z> )) <- <addressing_id>

(( <addressing_id> (( (( <+> || <-> )) <integer> {{ drop integers pull 1+ 2* }}
)) ?? )) <- <addressing_id_offset> ( n -- n' )

(( <[> {{ rot  }} <addressing_id_offset> {{ -rot }}
(( <,> {{ swap }} <addressing_id_offset> {{ swap }}
)) ?? <]> )) <- <offset> ( h v t -- h' v' t )

(( <r>   {{ 1 1  read_port }} (( <offset> )) ?? <data> {{ add_hv_string_port_type_vector }}
|| <w>   {{ 1 1 write_port }} (( <offset> )) ?? <data> {{ add_hv_string_port_type_vector }}
|| <NOP> {{ 1 1 dont_care_port 1 add_hv_port_type_vector }} )) <- <port_operation>

(( <port_operation> (( <:>
   <port_operation> {{ add_port_type_vector }} )) ** ))
<- <port_operation_list>

(( <port_operation_list> )) <- <operation>

(( <_1_dimensional_addressing_direction>
||            <row_addressing_direction>
||         <column_addressing_direction> )) <- <addressing_direction>

(( <addressing_direction>
((  (( <-> )) ?? <addressing_id>
(( <=> <->       <addressing_id>
)) ?? )) ?? )) <- <addressing_specifier>

(( <addressing_specifier> <(> <sub_march_element_list> <)>
|| <time_operation> {{ push_defaults }}
|| <operation> )) <- <sub_march_element>

(( <sub_march_element> (( <,>
   <sub_march_element> {{ set_port_type_vector }} )) **
)) <sub_march_element_list> rule

(( <tile_size> <(> <local_operation_list> <)> )) <- <tile_operation>

(( (( <addressing_id> )) ?? <(> <sub_march_element_list> <)>
|| <tile_operation> )) <- <tiled_sub_march_element>

(( <row_addressing_direction>
(( <addressing_id> )) ?? <(> <sub_march_element_list> <)>
|| <time_operation> {{ push_defaults }} )) <- <row_sub_march_element>

(( <row_sub_march_element> (( <,>
   <row_sub_march_element> {{ set_port_type_vector }} )) ** ))
<- <row_sub_march_element_list>

(( <row_addressing_direction> <tile_size> <(> <local_operation_list> <)>
)) <- <row_tile_operation>

(( (( <addressing_id> )) ?? <(> <row_sub_march_element_list> <)>
|| <row_tile_operation> )) <- <tiled_row_sub_march_element>

(( <column_addressing_direction>
(( <addressing_id> )) ?? <(> <sub_march_element_list> <)>
|| <time_operation> {{ push_defaults }} )) <- <column_sub_march_element>

(( <column_sub_march_element> (( <,>
   <column_sub_march_element> {{ set_port_type_vector }} )) ** ))
<- <column_sub_march_element_list>

(( <column_addressing_direction> <tile_size> <(> <local_operation_list> <)>
)) <- <column_tile_operation>

(( (( <addressing_id> )) ?? <(> <column_sub_march_element_list> <)>
|| <column_tile_operation> )) <- <tiled_column_sub_march_element>

(( <_1_dimensional_addressing_direction>        <tiled_sub_march_element>
||            <row_addressing_direction> <tiled_column_sub_march_element>
||         <column_addressing_direction>    <tiled_row_sub_march_element>
)) <- <normal_march_element>

(( <normal_march_element>
|| <global_march_element> {{ push_defaults }} ))
<- <march_element>

(( <march_element> (( <;>
   <march_element> {{ set_port_type_vector }} )) ** ))
<- <march_element_list>

(( <{> <march_element_list> <}> )) <- <test_algorithm>

(( <r> || <w> || <x> || <-> )) ?? <- <port_list>

(( <port_list> || <ALLPORTS> )) <- <test_type>

(( {{ mtla_init }} <test_type> <test_algorithm>
   {{ mtla_test_insert }}
)) <- <MTL_test>

( * ) set-current

<MTL_test> parser mtla

previous

: aparse ( str len )   init  true ?nextsym  mtla ;
