\ MTL Parser based on gforth's Gray version 5
\
\ Copyright 2006 Dennis Ruffer (druffer@worldnet.att.net)
\
\     This program is free software; you can redistribute it and/or modify
\     it under the terms of the GNU General Public License as published by
\     the Free Software Foundation; either version 2 of the License, or
\     (at your option) any later version.
\
\     This program is distributed in the hope that it will be useful,
\     but WITHOUT ANY WARRANTY; without even the implied warranty of
\     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
\     GNU General Public License for more details.
\
\     You should have received a copy of the GNU General Public License
\     along with this program; if not, write to the Free Software
\     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
\

require mtl_prs.fth
require Classes/Memory_chip.fth
require Data.fth

wordlist constant mtli_wordlist
get-current ( * ) mtli_wordlist set-current
get-order mtli_wordlist swap 1+ set-order

\ parsing stacks

20 stack integers
50 stack strings
30 stack row_offsets
30 stack column_offsets
10 stack numbers_of_rows
10 stack numbers_of_columns
10 stack addressing_directions
10 stack addressing_assignments
10 stack addressing_exclusions
10 stack addressing_indices
10 stack addressing_deltas
10 stack column_indices
10 stack column_deltas

: mtli_garbage_collection ( -- )
	chip_processes
	dup vsize 0 ?do  dup i [v]
		chip_operation_list
		dup vsize 0 ?do  dup i [v]
			chip_data vdelete
		loop  vdelete
	loop  vdelete
;

: add_next_operation ( operation -- chip_operation )
	chip_processing chip_operation_list @ if
		chip_processing chip_operation_list dup vsize 2dup 0 vinsert [v]
	else
		Chip_operation 1 vnew dup chip_processing chip_operation_list !
	then  swap over chip_core_operation c!
	0 over chip_data !
	0 over row_index !
	0 over row_address !
	0 over column_index !
	0 over column_address !
;

2variable data_background   0 0 data_background 2!

: add_chip_data ( chip_operation -- )   >r
	strings @ vsize mtli_memory_chip chips_word_width @ /mod
	swap if  1+  then  1 vnew r@ chip_data !
	strings @ vstring dup 1 > if
			base @ >r 2 base ! atod r> base !
	else  drop data_background 2@ rot c@
		dup [char] 1 = if  drop swap invert swap invert
		else  [char] 0 <> abort" Invalid data"  then
	then  r@ chip_data vstring dup 4 <= if  rot drop
	then  c!-le  strings pop  r> drop
;

: add_port_operation ( operation -- )
	add_next_operation dup >r add_chip_data
	row_offsets clear? if
		addressing_deltas pull? r@ row_address !
	else
		row_offsets pull r@ row_address !
	then
	column_offsets clear? if
		column_deltas pull? r@ column_address !
	else
		column_offsets pull r@ column_address !
	then
	addressing_indices dup clear? if  drop 0
	else  pull  dup 1- cells labeled_indices +
		@ 0< abort" label unefined"
	then  dup r@ row_index !  r@ column_index !
	column_indices dup clear? 0= if
		pull  dup 1- cells labeled_indices +
		@ 0< abort" label unefined"
		r@ column_index !
	else  drop  then  r> drop
;

: add_chip_process ( -- a )
	chip_processes @ if
		nest_operation add_next_operation drop
		chip_processes dup vsize 2dup 0 vinsert [v]  1 chip_process# +!
	else
		Chip_process 1 vnew dup chip_processes !  0 chip_process# !
	then
	0 over time_operation !
	0 over process_index !
	0 over 2_dimensions c!
	0 over swapped_dimensions c!
	0 over outer_backwards c!
	0 over inner_backwards c!
	0 over row_increment !
	0 over column_increment !
	0 over row_exclusion !
	0 over column_exclusion !
	0 over chip_operation_list !
;

: add_time_operation ( time -- )
	0 invert 1 rshift invert or		\ set msb
	add_chip_process
	time_operation !
;

: mtli_execute_chip_operation_list ( -- )
	traverse_chip 0= abort" traverse_chip had nothing to do!!!"
	mtli_garbage_collection
;

variable increment_rows		1 increment_rows !
variable increment_columns	1 increment_columns !

: tiled_rows    ( -- increment )   numbers_of_rows pull ;
: tiled_columns ( -- increment )   numbers_of_columns pull ;

: set_loop_parameters ( swapped_dimensions outer_backwards inner_backwards 2_dimensions -- )
	add_chip_process >r  r@ 2_dimensions c!
	r@ inner_backwards c!  r@ outer_backwards c!
	r@ swapped_dimensions c!
	increment_columns @ r@ column_increment !
	increment_rows @ r@ row_increment !
	0 r@ chip_operation_list !
	-1 r@ column_exclusion !
	-1 r@ row_exclusion !
	addressing_assignments dup clear? if  drop 0
	else  pull dup 1- cells labeled_indices +
		dup @ 0>= abort" label already used"
		chip_process# @ swap !
	then  r@ process_index !
	0 r> time_operation !
;

: do_double_loop ( swapped_dimensions outer_backwards inner_backwards -- )
	true set_loop_parameters
;

: column_top_left     ( -- )    true  true  true do_double_loop ;
: column_top_right    ( -- )    true  true false do_double_loop ;
: column_bottom_left  ( -- )    true false  true do_double_loop ;
: column_bottom_right ( -- )    true false false do_double_loop ;
:    row_top_left     ( -- )   false  true  true do_double_loop ;
:    row_top_right    ( -- )   false  true false do_double_loop ;
:    row_bottom_left  ( -- )   false false  true do_double_loop ;
:    row_bottom_right ( -- )   false false false do_double_loop ;

: do_single_loop ( swapped_dimensions outer_backwards -- )
	false false set_loop_parameters
;

:  forward_rows    ( -- )    true false do_single_loop ;
: backward_rows    ( -- )    true  true do_single_loop ;
:  forward_columns ( -- )   false false do_single_loop ;
: backward_columns ( -- )   false  true do_single_loop ;

: traverse_down ( -- )
	current_memory_type memory_address_order c@ column_first = if
		current_memory_type memory_row_order c@ top_first = if
			current_memory_type memory_column_order c@ left_first = if
				column_top_left  else  column_top_right
			then
		else  \ bottom_first
			current_memory_type memory_column_order c@ left_first = if
				column_bottom_left  else  column_bottom_right
			then
		then
	else  \ row_first
		current_memory_type memory_row_order c@ top_first = if
			current_memory_type memory_column_order c@ left_first = if
				row_top_left  else  row_top_right
			then
		else  \ bottom_first
			current_memory_type memory_column_order c@ left_first = if
				row_bottom_left  else  row_bottom_right
			then
		then
	then
;

: traverse_up ( -- )
	current_memory_type memory_address_order c@ column_first = if
		current_memory_type memory_row_order c@ top_first = if
			current_memory_type memory_column_order c@ left_first = if
				column_bottom_right  else  column_bottom_left
			then
		else  \ bottom_first
			current_memory_type memory_column_order c@ left_first = if
				column_top_right  else  column_top_left
			then
		then
	else  \ row_first
		current_memory_type memory_row_order c@ top_first = if
			current_memory_type memory_column_order c@ left_first = if
				row_bottom_right  else  row_bottom_left
			then
		else  \ bottom_first
			current_memory_type memory_column_order c@ left_first = if
				row_top_right  else  row_top_left
			then
		then
	then
;

: push_up_down_stack ( -- )
	addressing_directions pull case
		mtl_down of  traverse_down  endof
		mtl_up of  traverse_up  endof
		dup of
			true abort" MTL internal error: 1_dimensional_addressing_direction not recognized"
		endof
	endcase
;

: push_north_south_stack ( -- )
	addressing_directions pull case
		mtl_south of  forward_rows  endof
		mtl_north of  backward_rows  endof
		dup of
			true abort" MTL internal error: row_addressing_direction not recognized"
		endof
	endcase
;

: push_east_west_stack ( -- )
	addressing_directions pull case
		mtl_east of  forward_columns  endof
		mtl_west of  backward_columns  endof
		dup of
			true abort" MTL internal error: column_addressing_direction not recognized"
		endof
	endcase
;

: push_north_south_east_west_stack ( -- )
	addressing_directions pull case
		mtl_north of
			addressing_directions pull case
				mtl_east of  column_top_right  endof
				mtl_west of  column_top_left  endof
				dup of
					true abort" MTL internal error: column_addressing_direction not recognized"
				endof
			endcase
		endof
		mtl_south of
			addressing_directions pull case
				mtl_east of  column_bottom_right  endof
				mtl_west of  column_bottom_left  endof
				dup of
					true abort" MTL internal error: column_addressing_direction not recognized"
				endof
			endcase
		endof
		dup of
			true abort" MTL internal error: row_addressing_direction not recognized"
		endof
	endcase
;

: push_east_west_north_south_stack ( -- )
	addressing_directions pull case
		mtl_east of
			addressing_directions pull case
				mtl_north of  row_bottom_left  endof
				mtl_south of  row_bottom_right  endof
				dup of
					true abort" MTL internal error: row_addressing_direction not recognized"
				endof
			endcase
		endof
		mtl_west of
			addressing_directions pull case
				mtl_north of  row_top_left  endof
				mtl_south of  row_top_right  endof
				dup of
					true abort" MTL internal error: row_addressing_direction not recognized"
				endof
			endcase
		endof
		dup of
			true abort" MTL internal error: column_addressing_direction not recognized"
		endof
	endcase
;

: push_all_stack ( -- )
	addressing_directions pull case
		mtl_down of  traverse_down  endof
		mtl_up of  traverse_up  endof
		mtl_south of  forward_rows  endof
		mtl_north of  backward_rows  endof
		mtl_east of  forward_columns  endof
		mtl_west of  backward_columns  endof
		dup of
			true abort" MTL internal error: addressing_direction not recognized"
		endof
	endcase
;

: mtli_init ( -- )
	labeled_indices 26 cells -1 fill
	1 increment_columns !
	1 increment_rows !
	integers clear
	row_offsets clear
	column_offsets clear
	numbers_of_rows clear
	numbers_of_columns clear
	addressing_directions clear
	addressing_assignments clear
	addressing_exclusions clear
	addressing_indices clear
	addressing_deltas clear
	column_indices clear
	column_deltas clear
	begin  strings clear? 0= while
		strings dup @ vdelete pop
	repeat
;

: mtli_test_check ( -- )
	integers clear? 0= abort" integers left"
	strings clear? 0= abort" strings left"
	row_offsets clear? 0= abort" row_offsets left"
	column_offsets clear? 0= abort" column_offsets left"
	numbers_of_rows clear? 0= abort" numbers_of_rows left"
	numbers_of_columns clear? 0= abort" numbers_of_columns left"
	addressing_directions clear? 0= abort" addressing_directions left"
	addressing_assignments clear? 0= abort" addressing_assignments left"
	addressing_exclusions clear? 0= abort" addressing_exclusions left"
	addressing_indices clear? 0= abort" addressing_indices left"
	addressing_deltas clear? 0= abort" addressing_deltas left"
	column_indices clear? 0= abort" column_indices left"
	column_deltas clear? 0= abort" column_deltas left"
	false mtli_fault_found !
;

\ non-terminals

nonterminal <sub_march_element_list>

\ parser

(( <unsigned_integer> {{ ident 2@ atoi integers push }} )) <- <integer>
(( <unsigned_integer> {{ ident 2@ atos strings push }} )) <- <data>

(( <t> {{ -1 }}
(( <integer>  {{ drop integers pull }} )) ??
{{ add_time_operation }} )) <- <time_operation>

(( <time_operation> )) <- <global_march_element>

(( <integer> {{ integers pull row_offsets push }} )) <- <row_offset>
(( <integer> {{ integers pull column_offsets push }} )) <- <column_offset>

(( <[> <row_offset> <,> <column_offset> <]> )) <- <location>

(( <r> <location> <data> {{  read_operation add_port_operation }}
|| <w> <location> <data> {{ write_operation add_port_operation }}
)) <- <local_operation>

(( <integer> {{ integers pull numbers_of_rows push }} )) <- <number_of_rows>
(( <integer> {{ integers pull numbers_of_columns push }} )) <- <number_of_columns>

(( <[> <number_of_rows> <,> <number_of_columns> <]> )) <- <tile_size>

(( <local_operation> (( <,> <local_operation> )) ** )) <- <local_operation_list>

(( <UPDOWN> {{ mtl_down addressing_directions push }}
|| <DOWN>   {{ mtl_down addressing_directions push }}
|| <UP>     {{ mtl_up   addressing_directions push }}
)) <- <_1_dimensional_addressing_direction>

(( <EASTWEST> {{ mtl_east addressing_directions push }}
|| <EAST>     {{ mtl_east addressing_directions push }}
|| <WEST>     {{ mtl_west addressing_directions push }}
)) <- <row_addressing_direction>

(( <NORTHSOUTH> {{ mtl_south addressing_directions push }}
|| <SOUTH>      {{ mtl_south addressing_directions push }}
|| <NORTH>      {{ mtl_north addressing_directions push }}
)) <- <column_addressing_direction>

(( <a> {{  1 addressing_assignments push }}
|| <b> {{  2 addressing_assignments push }}
|| <c> {{  3 addressing_assignments push }}
|| <d> {{  4 addressing_assignments push }}
|| <e> {{  5 addressing_assignments push }}
|| <f> {{  6 addressing_assignments push }}
|| <g> {{  7 addressing_assignments push }}
|| <h> {{  8 addressing_assignments push }}
|| <i> {{  9 addressing_assignments push }}
|| <j> {{ 10 addressing_assignments push }}
|| <k> {{ 11 addressing_assignments push }}
|| <l> {{ 12 addressing_assignments push }}
|| <m> {{ 13 addressing_assignments push }}
|| <n> {{ 14 addressing_assignments push }}
|| <o> {{ 15 addressing_assignments push }}
|| <p> {{ 16 addressing_assignments push }}
|| <q> {{ 17 addressing_assignments push }}
|| <r> {{ 18 addressing_assignments push }}
|| <s> {{ 19 addressing_assignments push }}
|| <t> {{ 20 addressing_assignments push }}
|| <u> {{ 21 addressing_assignments push }}
|| <v> {{ 22 addressing_assignments push }}
|| <w> {{ 23 addressing_assignments push }}
|| <x> {{ 24 addressing_assignments push }}
|| <y> {{ 25 addressing_assignments push }}
|| <z> {{ 26 addressing_assignments push }}
)) <- <addressing_id>

(( <addressing_id> {{ addressing_assignments pull
					  addressing_indices push  0 addressing_deltas push }}
	  (( (( <+>    {{  1 addressing_deltas dup pop push }}
	     || <->    {{ -1 addressing_deltas dup pop push }}
	     )) <integer> {{ addressing_deltas pull  integers pull *
	     				 addressing_deltas push }}
      )) ?? )) <- <addressing_id_offset>

(( <[> <addressing_id_offset>
(( <,> <addressing_id_offset> {{ addressing_indices pull  column_indices push
								 addressing_deltas pull  column_deltas push }}
)) ?? <]> )) <- <offset>

(( <r> (( <offset> )) ?? <data> {{  read_operation add_port_operation }}
|| <w> (( <offset> )) ?? <data> {{ write_operation add_port_operation }}
|| <NOP> {{ no_operation add_port_operation }} )) <- <port_operation>

(( <port_operation> (( <:>
   <port_operation> )) ** ))
<- <port_operation_list>

(( <port_operation_list> )) <- <operation>

(( <_1_dimensional_addressing_direction>
||            <row_addressing_direction>
||         <column_addressing_direction> )) <- <addressing_direction>

(( <addressing_direction>
((  (( <-> )) ?? <addressing_id> {{ addressing_assignments pull drop }} \ how is this different than normal nesting???
(( <=> <->       <addressing_id> {{ addressing_assignments pull
								    addressing_exclusions push }}
)) ?? )) ?? )) <- <addressing_specifier>

(( <addressing_specifier> <(>
   {{ push_all_stack }} <sub_march_element_list> <)>
   {{ -1 chip_process# +! }}
|| <time_operation> || <operation> )) <- <sub_march_element>

(( <sub_march_element> (( <,>
   <sub_march_element> )) **
)) <sub_march_element_list> rule

(( <tile_size> <(>
   {{ tiled_columns increment_columns !
	  tiled_rows increment_rows !
	  push_up_down_stack
	}} <local_operation_list> <)>
	{{ -1 chip_process# +! }}
)) <- <tile_operation>

(( <row_addressing_direction> <tile_size> <(>
   {{ tiled_columns increment_columns !
	  tiled_rows increment_rows !
	  push_north_south_east_west_stack
	}} <local_operation_list> <)>
	{{ -1 chip_process# +! }}
)) <- <row_tile_operation>

(( <column_addressing_direction> <tile_size> <(>
   {{ tiled_columns increment_columns !
	  tiled_rows increment_rows !
	  push_east_west_north_south_stack
	}} <local_operation_list> <)>
	{{ -1 chip_process# +! }}
)) <- <column_tile_operation>

(( (( <addressing_id> )) ?? <(>
   {{ push_up_down_stack }} <sub_march_element_list> <)>
   {{ -1 chip_process# +! }}
|| <tile_operation> )) <- <tiled_sub_march_element>

(( <row_addressing_direction>
   (( <addressing_id> )) ?? <(>
   {{ push_north_south_stack }} <sub_march_element_list> <)>
   {{ -1 chip_process# +! }}
|| <time_operation> )) <- <row_sub_march_element>

(( <row_sub_march_element> (( <,>
   <row_sub_march_element> )) ** ))
<- <row_sub_march_element_list>

(( (( <addressing_id> )) ?? <(>
   {{ push_north_south_stack }} <row_sub_march_element_list> <)>
   {{ -1 chip_process# +! }}
|| <row_tile_operation> )) <- <tiled_row_sub_march_element>

(( <column_addressing_direction>
   (( <addressing_id> )) ?? <(>
   {{ push_east_west_stack }} <sub_march_element_list> <)>
   {{ -1 chip_process# +! }}
|| <time_operation> )) <- <column_sub_march_element>

(( <column_sub_march_element> (( <,>
   <column_sub_march_element> )) ** ))
<- <column_sub_march_element_list>

(( (( <addressing_id> )) ?? <(>
   {{ push_east_west_stack }} <column_sub_march_element_list> <)>
   {{ -1 chip_process# +! }}
|| <column_tile_operation> )) <- <tiled_column_sub_march_element>

(( <_1_dimensional_addressing_direction>        <tiled_sub_march_element>
||            <row_addressing_direction> <tiled_column_sub_march_element>
||         <column_addressing_direction>    <tiled_row_sub_march_element>
)) <- <normal_march_element>

(( <normal_march_element> {{ mtli_execute_chip_operation_list }}
|| <global_march_element> {{ mtli_execute_chip_operation_list }}
)) <- <march_element>

(( <march_element> (( <;>
   <march_element> )) ** ))
<- <march_element_list>

(( <{> {{ mtli_garbage_collection }}
   <march_element_list> <}> {{ mtli_garbage_collection }}
)) <- <test_algorithm>

(( <r> || <w> || <x> || <-> )) ?? <- <port_list>

(( <port_list> || <ALLPORTS> )) <- <test_type>

(( {{ mtli_init }} <test_type> <test_algorithm>
   {{ mtli_test_check }}
)) <- <MTL_test>

( * ) set-current

<MTL_test> parser mtli

previous

: iparse ( str len )   init  true ?nextsym  mtli ;
